#include <map>
#include <unordered_set>
#include <cstdint>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <algorithm>
#include <set>
#include <bitset>
#include <unistd.h>
#include <chrono>
#include <stdio.h>
#include "lzsdd.hpp"

using namespace std;

#define INTER -1

typedef uint64_t my_uint;

typedef long long int int64; // 64ビット整数型

// 最初のDecompositionノードを作成して返す。

inline LZsdd* CreateRootNode(const int number_of_vertices, const int cur, const vector<my_vec> *frontier){
	LZsdd *node = new LZsdd;	
	node->mate.resize(number_of_vertices);
	node->id = cur;
	node->edge = 0;		
	// deg, comp を初期化
	for (int i = 0; i < number_of_vertices; ++i){
		node->mate[i] = i;
	}
	node->frontier = (vector<my_vec>*)frontier;
	return node;
}




LZsdd* MergeMate(const LZsdd* n_hat, const LZsdd* n_hat2, const State* state, const int cur,
				 const vector<my_vec>& uni, const vector<my_vec>& inter, int max_vertex)
{
	vector<my_vec> l_mate(max_vertex+5);
	vector<my_vec> r_mate(max_vertex+5);
	vector<my_vec> opposite(max_vertex+5);
	vector<my_vec> opposite2(max_vertex+5);
	
	for(int i = 0; i<uni.size(); i++){
		int v=uni[i];
		l_mate[v] = n_hat->mate[v];
		r_mate[v] = n_hat2->mate[v];
		opposite[v] = 0;
		opposite2[v] = 1;
	}
	int s=state->start, t=state->end;
	l_mate[s] = n_hat->mate[s];
	l_mate[t] = n_hat->mate[t];
	r_mate[s] = n_hat2->mate[s];
	r_mate[t] = n_hat2->mate[t];
	opposite[s] = opposite[t] = 0;
	opposite2[s] = opposite2[t] = 1;	

	LZsdd *n_prime = CreateRootNode(state->max_vertex, cur, (vector<my_vec>*)&state->vnodes[cur].F);

	int v;//ここではフロンティアの各頂点をvに
	for (unsigned int i = 0; i < inter.size(); ++i) { // 左右のintersectionについて比較			
		v = inter[i];
		//片方次数0
		if(l_mate[v]==v || r_mate[v]==v){
			;//do nothing
		}
		//片方次数2　他方は1以上 s,tでない
		else if((l_mate[v]==INTER || r_mate[v]==INTER)){
			//合計次数３以上となる
			delete n_prime;
			n_prime=NULL;
			goto _exit_merge;
		}
		//両方次数1
		else{
			//(3) vとwが同じ連結成分 サイクルが生じる
			if(l_mate[v]==r_mate[v]){
				delete n_prime;
				n_prime = NULL;
				goto _exit_merge;
			}
			//(1) sまたはtに辺が2本以上接続する．
			else if(v == s || v == t){
				delete n_prime;
				n_prime = NULL;
				goto _exit_merge;
			}
			else{
				int vl0 = l_mate[v];
				int vr0 = r_mate[v];
				int vl1 = opposite[v];
				int vr1 = opposite2[v];
				l_mate[v] = INTER;
				r_mate[v] = INTER;
				if(vl1 == 0){//左側
					l_mate[vl0] = vr0;
					n_prime->mate[vl0] = vr0;
					opposite[vl0] = vr1;
				}
				else{
					r_mate[vl0] = vr0;
					n_prime->mate[vl0] = vr0;
					opposite2[vl0] = vr1;
				}
				if(vr1 == 0){//左側
					l_mate[vr0] = vl0;
					n_prime->mate[vr0] = vl0;
					opposite[vr0] = vl1;
				}
				else{
					r_mate[vr0] = vl0;
					n_prime->mate[vr0] = vl0;
					opposite2[vr0] = vl1;
				}
				
			}
		}
		//ShowState(n2_hat, "n2_hat");
		//ShowState(n2_hat2, "n2_hat2");
	}
	//mateを振り直す
	//s,tは更新
	if(l_mate[s]!=s)
		n_prime->mate[s] = l_mate[s];
	else
		n_prime->mate[s] = r_mate[s];
	if(n_prime->mate[s]==INTER){
		delete n_prime;
		n_prime = NULL;
		goto _exit_merge;
	}

	if(l_mate[t]!=t)
		n_prime->mate[t] = l_mate[t];
	else
		n_prime->mate[t] = r_mate[t];
	if(n_prime->mate[t]==INTER){
		delete n_prime;
		n_prime = NULL;
		goto _exit_merge;
	}

	for (unsigned int i = 0; i < uni.size(); i++) { // 左右のunionについて設定			
		v = uni[i];
		//片方しか次数1でないため
		if(l_mate[v]!=v)
			n_prime->mate[v] = l_mate[v];
		else
			n_prime->mate[v] = r_mate[v];
	}
	
 _exit_merge:
	return n_prime;
}



void GetDecompVec(const int v_l, const int v_r, const int v, const State *state,
				  vector<my_vec> &diff, vector<my_vec> &uni, vector<my_vec> &inter){
	const vector<my_vec> &left = state->vnodes[v_l].F;
	const vector<my_vec> &right = state->vnodes[v_r].F;
	const vector<my_vec> &center = state->vnodes[v].F;
	
	set_union(left.cbegin(), left.cend(),
		right.cbegin(), right.cend(),
		back_inserter(uni));
	set_intersection(left.cbegin(), left.cend(),
		right.cbegin(), right.cend(),
		back_inserter(inter));
	set_difference(uni.cbegin(), uni.cend(),
		center.cbegin(), center.cend(),
		back_inserter(diff));

}


//Simple Path固有処理


//Simple Path固有処理
LZsdd* DecompChild(const LZsdd* n_hat, const LZsdd* n_hat2, const State* state, const int v,
					const vector<my_vec> &diff, const vector<my_vec> &uni, const vector<my_vec> &inter, bool& endflg){
	bool flg=true;
	
	//フロンティアの処理
	//左右mateマージ (1),(2),(3)も判定
	const int s = state->start;
	const int t = state->end;

	LZsdd* n_prime = MergeMate(n_hat, n_hat2, state, v, uni, inter, state->max_vertex);

	if(n_prime != NULL){
		//余りの処理
		for(const auto i : diff){
			//(4) s または t の次数が 0 に確定する
			if((i == s || i == t) && n_prime->mate[i]==i){
				flg=false;
				delete n_prime;
				goto _exit_decomp;
			}
			//(5) s − t 経路が途中で分断される
			if((i != s && i != t) && n_prime->mate[i]!=INTER && n_prime->mate[i] != i){
				flg=false;
				delete n_prime;
				goto _exit_decomp;
			}
		}
	    if(n_prime->mate[s]==t && n_prime->mate[t]==s){
			endflg = true;
			for (unsigned int i = 0; i < uni.size(); i++) {
				int v = uni[i];
				if(v != s && v != t){
					flg &= (n_prime->mate[v]==INTER || n_prime->mate[v]==v);
					if(flg==false){
						delete n_prime;
						goto _exit_decomp;
					}
					endflg &= (n_prime->mate[v]==INTER || n_prime->mate[v]==v);
				}
			}
        }
	}
	else{
		flg = false;
	}
 _exit_decomp:
	if(flg==false){
		return NULL;
	}
	else{
		return n_prime;
	}
}



inline void LeafChild(const int v, const int e, vector<my_set>& N, const State *state){
	//False
	LZsdd *node1 = CreateRootNode(state->max_vertex, v, (vector<my_vec>*)&state->vnodes[v].F);
	const Edge &edge = state->edges[e];
	node1->edge = OneTerminal;
#ifdef  SET_VEC
	vector<my_vec> tmp;
	node1->ans.emplace_back(move(tmp));	
#endif
	N[v].emplace(node1);
	//True
	LZsdd *node2 = CreateRootNode(state->max_vertex, v, (vector<my_vec>*)&state->vnodes[v].F);
	//mate更新
	int tsrc = node2->mate[edge.src];
	int tdest = node2->mate[edge.dest];
	node2->mate[tsrc] = tdest; node2->mate[tdest] = tsrc;
	
	node2->edge = e;
#ifdef  SET_VEC
	vector<my_vec> t;
	t.emplace_back(e);
	node2->ans.emplace_back(move(t));
#endif	
	N[v].emplace(node2);
}

static void usage(char *argv[]){
	fprintf(stderr,"Usage: %s [-g .] [-v .]\n",argv[0]);
	fprintf(stderr," -g FILE   : input graph\n");
	fprintf(stderr," -v FILE   : vtree of the graph\n");
	exit(1);
}

State state;

int main(int argc, char *argv[]) {


	ReadGraph(stdin, &state);
	//ReadVtree(vin, &state);
	//printf("gen vtree\n");
	BranchTree bt;
	bt.initialize(&state);
	bt.decomp_graph(&state);
    bt.gen_vtree(0, &state);
	//printf("gen index\n");
	Dfrontier(&state);
	return 0;
}


